# -*- coding: utf-8 -*-

# Standard Library
import json
import datetime
import zlib
import base64
import socket

# Vendor Library
from celery import group
from django.conf import settings
from django.template import loader
from django.utils.datastructures import SortedDict
import suds
from suds.client import Client
from suds.sax.element import Element

# Project Library
from kinder.core.async.tasks import celery_app
from kinder.core.dict.models import (
    DouType, HNE, GroupCountNorm, UnitKind)
from kinder.core.group.models import Group, Pupil, GroupStatusEnum
from kinder.core.emie_unit.models import UnitDocFiles, DopService
from gisdo.utils import UnitHelper
from gisdo.alchemy_session import session_scope
from gisdo.index import (
    ApplicationIndex, QueueIndex, EnrolledIndex, CapacitiesIndex)
from gisdo import constants as const
from gisdo.models import (
    ReportFormRow, ReportForm, ScheduleSettings,
    InstallationType)
from gisdo.queries import QueueCacheBuilder
from gisdo import settings as gs_settings
from gisdo.exceptions import (
    SendReportException, DocumentCreateException,
    XMLRenderException, SOAPClientCreateException,
    NoDataInMo)
from gisdo.logger import Logger

class ReportCacheMixin(object):
    APPLICATIONS = 'applications'
    QUEUES = 'queues'
    ENROLLED = 'enrolled'
    CAPACITIES = 'capacities'

    DATA_TYPE = [
        APPLICATIONS,
        QUEUES,
        ENROLLED,
        CAPACITIES
    ]

    def __init__(self):

        self._cache = {
            self.APPLICATIONS: {},
            self.QUEUES: {},
            self.ENROLLED: {},
            self.CAPACITIES: {}
        }

    def _update_cache(self, data, data_type):
        """
        Обновление кэша
        """

        for key, value in data.iteritems():
            if not key in self._cache[data_type]:
                self._cache[data_type][key] = value
            elif isinstance(value, dict):
                for age_cat in data[key]:
                    self._cache[
                        data_type][key][age_cat] += data[key][age_cat]


class ReportService(ReportCacheMixin):
    def __init__(
            self, main_unit,
            set_progress_func=None,
            user=None):

        super(ReportService, self).__init__()

        self._unit_helper = UnitHelper(main_unit)
        self._report_form = ReportForm(
            user=user, unit=main_unit,
            in_progress=True)
        self._set_progress = set_progress_func or (lambda x: x)
        self._schedule_settings = ScheduleSettings.get_settings()

    def collect(self, send=True):
        self._report_form.save()

        # Счетчик кол-ва ошибок при отправке
        send_error_count = 0

        # На этом уровне перебираем учреждения типа МО
        current_mo_num = 1
        mo_count = len(list(self._unit_helper.get_report_units()))

        # Создаем группу параллельно выполняющихся подзадач
        # Каждая задача отвечает за подсчет показателей отдельного МО
        collect_data_tasks = group(
            self._collect_data.s(
                mo, self._unit_helper._unit, self._report_form
            ) for mo in self._unit_helper.get_report_units())

        collect_data_tasks_results = collect_data_tasks(
            queue=gs_settings.CELERY_WORKER_TASK_QUEUE)

        mo_for_send = []

        # Обрабатываем результаты завершившихся задач
        for (mo, task_result, exception) in collect_data_tasks_results.iterate():
            self._set_progress(
                u'Обработано МО: %s из %s' % (
                    current_mo_num, mo_count))

            try:
                # Выбрасываем отложенное исключение
                if exception:
                    raise exception
            except Exception as e:
                self._set_progress(
                    values={
                        u'Во время сбора данных по'
                        u' %s произошла ошибка' % mo.name:
                        e.message
                    })
                current_mo_num += 1
                continue
            else:
                self._update_data(task_result)

            self._set_progress(
                values={
                    u'Данные по %s успешно собраны' % mo.name: u''
                })

            current_mo_num += 1

            if send:
                mo_for_send.append(mo)

        current_mo_num = 1

        if send:
            self._set_progress(u'Подготовка к отправке данные МО')

            # Создаем группу параллельно выполняющихся подзадач
            # Каждая задача отвечает за отправку показателей отдельного МО
            send_data_tasks = group(
                self.send_mo_data.s(
                    mo, self._unit_helper._unit,
                    self._report_form, self._schedule_settings
                ) for mo in mo_for_send)

            send_data_tasks_results = send_data_tasks(
                queue=gs_settings.CELERY_WORKER_TASK_QUEUE)

            for (mo, exception) in send_data_tasks_results.iterate():
                self._set_progress(
                    u'Отправлены данные МО: %s из %s' % (
                        current_mo_num, mo_count))

                try:
                    if exception:
                        raise exception
                except Exception as e:
                    self._set_progress(
                        values={
                            u'Произошла ошибка при '
                            u'передаче данных по %s' % mo.name:
                            e.message
                        })
                    current_mo_num += 1
                    send_error_count += 1
                    continue

                self._set_progress(
                    values={
                        u'Данные по %s успешно отправлены' % mo.name:
                        u''
                    })

                current_mo_num += 1

        if self._unit_helper.is_region:
            ReportFormRow.objects.get_or_create(
                report=self._report_form,
                unit=self._unit_helper.get_main_unit(),
                applications=json.dumps(self._cache[self.APPLICATIONS]),
                queues=json.dumps(self._cache[self.QUEUES]),
                enrolled=json.dumps(self._cache[self.ENROLLED]),
                capacities=json.dumps(self._cache[self.CAPACITIES]))

        self._report_form.in_progress = False

        # Если все пакеты были отправлены то считаем что отчет отправлен
        if send and send_error_count == 0:
            self._report_form.sent_time = datetime.datetime.now()

        self._report_form.date = datetime.datetime.now()
        self._report_form.save()

    @celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
    def _collect_data(mo, main_report_unit, report_form):
        unit_helper = UnitHelper(main_report_unit)

        data = None
        exception = None

        try:
            data = DataCollector(
                mo, unit_helper).collect(report_form)
        # Отложенное исключение
        except Exception as e:
            exception = e
        return (mo, data, exception)

    def _update_data(self, mo_data):
        if self._unit_helper.is_region:
            self._update_cache(mo_data[self.APPLICATIONS], self.APPLICATIONS)
            self._update_cache(mo_data[self.QUEUES], self.QUEUES)
            self._update_cache(mo_data[self.ENROLLED], self.ENROLLED)
            self._update_cache(mo_data[self.CAPACITIES], self.CAPACITIES)

    @celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
    def send_mo_data(mo, main_report_unit, report_form, settings):
        unit_helper = UnitHelper(main_report_unit)

        exception = None
        try:
            Transmitter(
                report_form,
                settings,
                unit_helper).send(mo)
        except Exception as e:
            exception = e

        return (mo, exception)


class Transmitter(object):

    def __init__(self, report_form, schedule_settings, unit_helper):

        self._report_form = report_form
        self._schedule_settings = schedule_settings
        self._unit_helper = unit_helper

    def send(self, unit=None):

        if not unit:
            unit = self._unit_helper.get_main_unit()
        elif unit.kind.id not in [UnitKind.MO, UnitKind.REGION]:
            raise SendReportException()

        # Список МО по которым не удалось отправить данные
        not_send_mo = []
        for mo in unit.get_descendants(
                include_self=True).filter(kind__id=UnitKind.MO):

            try:
                document = DocumentService(
                    mo, self._report_form.date,
                    self._schedule_settings,
                    self._unit_helper,
                    self._report_form).create()
            except NoDataInMo as e:
                not_send_mo.append(mo.name)
                Logger.add_record(e.message, level=Logger.WARNING)
                continue
            except Exception as e:
                Logger.add_record(e.message, level=Logger.ERROR)
                raise DocumentCreateException()

            try:
                xml_content = TemplateLoader(
                    self._schedule_settings).load(document)
            except Exception as e:
                Logger.add_record(e.message, level=Logger.ERROR)
                raise XMLRenderException()

            BroadCaster(self._schedule_settings).send(
                xml_content)

        return not_send_mo


class DataCollector(ReportCacheMixin):
    def __init__(self, mo, unit_helper):

        super(DataCollector, self).__init__()

        self._mo = mo
        self._unit_helper = unit_helper
        self._queue_cache = QueueCacheBuilder.build_cache(self._mo)

    def collect(self, report_form):

        # Признак наличия садов по которым можно
        # собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo()

        # Собираем по ДОУ
        for dou in dou_list:

            self._dou_collect(report_form, dou)

        # Данные по МО лежат в кэше
        ReportFormRow.objects.get_or_create(
            applications=json.dumps(self._cache[self.APPLICATIONS]),
            queues=json.dumps(self._cache[self.QUEUES]),
            enrolled=json.dumps(self._cache[self.ENROLLED]),
            capacities=json.dumps(self._cache[self.CAPACITIES]),
            report=report_form,
            unit=self._mo)

        return self._cache

    def _dou_collect(self, report_form, dou):

        applications = self._collect_applications_data(dou)
        queues = self._collect_queue_data(
            dou, queue_cache=self._queue_cache)
        enrolled = self._collect_enrolled_data(dou, queues)
        capacities = self._collect_capacities(dou)

        ReportFormRow.objects.get_or_create(
            report=report_form, unit=dou,
            applications=json.dumps(applications),
            queues=json.dumps(queues),
            enrolled=json.dumps(enrolled),
            capacities=json.dumps(capacities))

    def _collect_applications_data(self, dou):

        # Словарь с id-шниками показателей
        applications = SortedDict([
            ('1', SortedDict()),
            ('1.1', SortedDict()),
            ('2.1', SortedDict()),
            ('3.1', SortedDict()),
            ('4', SortedDict()),
            ('4.1', SortedDict()),
            ('5', SortedDict()),
            ('5.1', SortedDict()),
            ('6', SortedDict()),
            ('6.1', SortedDict()),
            ('36', SortedDict())

        ])

        with session_scope() as session:
            index = ApplicationIndex(dou, session)

            for value in const.AGE_CATEGORIES_FULL.values():
                # Общее количество заявлений на постановку в очередь
                applications['1'][value] = index.get_count(
                    age_range=value)

                applications['2.1'][value] = index.get_count(
                    index_type=ApplicationIndex.COUNT_BY_DELIVERY, portal=1,
                    age_range=value, on_date=datetime.date.today)
                applications['3.1'][value] = index.get_count(
                    index_type=ApplicationIndex.COUNT_BY_DELIVERY,
                    age_range=value, on_date=datetime.date.today)

                applications['1.1'][value] = (
                    applications['2.1'][value] +
                    applications['3.1'][value])

                # Общее количество заявлений на постановку в очередь
                # в негосударственные ДОО
                applications['4'][value] = index.get_count(
                    dou_type=const.NOT_GOVERNMENT, age_range=value)
                applications['4.1'][value] = index.get_count(
                    dou_type=const.NOT_GOVERNMENT, age_range=value,
                    on_date=datetime.date.today)

                # Общее количество заявлений на постановку в очередь
                # у ИП
                applications['5'][value] = index.get_count(
                    dou_type=const.IP, age_range=value)
                applications['5.1'][value] = index.get_count(
                    dou_type=const.IP, age_range=value,
                    on_date=datetime.date.today)

                # Общее количество заявлений на постановку в очередь
                # в группы кратковременного пребывания
                applications['6.1'][value] = index.get_count(
                    index_type=ApplicationIndex.COUNT_BY_GROUP_TYPE,
                    age_range=value, on_date=datetime.date.today)

            for value in const.AGE_CATEGORIES_CUT.values():
                # Общее количество заявлений на постановку в очередь
                # в группы кратковременного пребывания
                applications['6'][value] = index.get_count(
                    index_type=ApplicationIndex.COUNT_BY_GROUP_TYPE,
                    age_range=value)

            applications['36'][const.ALL] = index.get_count(
                index_type=ApplicationIndex.COUNT_DENIED,
                on_date=datetime.date.today)

            self._update_cache(applications, self.APPLICATIONS)

            return applications

    def _collect_queue_data(self, dou, queue_cache):

        queues = SortedDict([
            ('7', SortedDict()), ('7.1', SortedDict()),
            ('7.2', SortedDict()), ('7.3', SortedDict()),
            ('8', SortedDict()), ('8.1', SortedDict()),
            ('9', SortedDict()), ('9.1', SortedDict()),
            ('10', SortedDict()), ('11', SortedDict()),
            ('12', SortedDict()), ('12.1', SortedDict()),
            ('13.1', SortedDict()), ('14.1', SortedDict()),
            ('15', SortedDict()), ('15.1', SortedDict()),
            ('16', SortedDict()), ('16.1', SortedDict()),
            ('17', SortedDict()), ('17.1', SortedDict()),
            ('18', SortedDict()), ('18.1', SortedDict()),
            ('37', SortedDict()), ('38', SortedDict()),
            ('39', SortedDict()), ('40', SortedDict()),
            ('41', SortedDict()), ('42', SortedDict())
        ])

        with session_scope() as session:

            application_index = ApplicationIndex(unit=dou, session=session)
            queue_index = QueueIndex(unit=dou, queue_cache=queue_cache)

            for value in const.AGE_CATEGORIES_FULL.values():
                # Отложенный спрос
                queues['7'][value] = queue_index.get_count(
                    index_type=QueueIndex.DEFERRED, age_range=value)
                queues['7.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.DEFERRED,
                    age_range=value, on_date='CURRENT')
                queues['7.2'][value] = queue_index.get_count(
                    index_type=QueueIndex.QUEUE_IN_NEXT_YEAR,
                    age_range=value, on_date='CURRENT')
                queues['7.3'][value] = queue_index.get_count(
                    index_type=QueueIndex.DEFERRED_QUEUE_IN_NEXT_YEAR,
                    age_range=value, on_date='CURRENT')

                # Актуальный спрос
                queues['18'][value] = queue_index.get_count(
                    index_type=QueueIndex.ACTUAL_CALL, age_range=value)
                queues['18.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.ACTUAL_CALL, age_range=value,
                    on_date='CURRENT')

                # Актуальный спрос + Желает сменить ДОО
                queues['8'][value] = (
                    queues['18'][value] + application_index.get_count(
                        index_type=ApplicationIndex.WANT_CHANGE_DOU,
                        age_range=value))
                queues['8.1'][value] = (
                    queues['18.1'][value] + application_index.get_count(
                        index_type=ApplicationIndex.WANT_CHANGE_DOU,
                        age_range=value, on_date=datetime.date.today))

                # Количество детей в очереди, имеющих право
                # на внеочередное и первоочередное зачисление
                queues['9'][value] = queue_index.get_count(
                    index_type=QueueIndex.PRIVILEGE, age_range=value)
                queues['9.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.PRIVILEGE,
                    age_range=value, on_date='CURRENT')

                # Количество детей в очереди с ограниченными
                # возможностями по здоровью
                queues['12'][value] = queue_index.get_count(
                    index_type=QueueIndex.DISABLED, age_range=value)
                queues['12.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.DISABLED,
                    age_range=value, on_date='CURRENT')

                # Количество детей в очередь в негос. ДОУ
                queues['15'][value] = queue_index.get_count(
                    index_type=QueueIndex.ACTUAL_CALL, age_range=value,
                    dou_type=const.NOT_GOVERNMENT)
                queues['15.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.ACTUAL_CALL, age_range=value,
                    dou_type=const.NOT_GOVERNMENT, on_date='CURRENT')

                # Количество детей в очередь в частные ДОУ
                queues['16'][value] = queue_index.get_count(
                    index_type=QueueIndex.ACTUAL_CALL, age_range=value,
                    dou_type=const.IP)
                queues['16.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.ACTUAL_CALL, age_range=value,
                    dou_type=const.IP, on_date='CURRENT')

                queues['37'][value] = queue_index.get_count(
                    index_type=QueueIndex.CURRENT_DATE_ACTUAL_CALL,
                    age_range=value, on_date='CURRENT')

                queues['38'][value] = queue_index.get_count(
                    index_type=QueueIndex.QUEUE_IN_THIS_YEAR,
                    age_range=value, on_date='CURRENT')

                queues['39'][value] = queue_index.get_count(
                    index_type=QueueIndex.QUEUE_IN_CURRENT_CALENDAR_YEAR,
                    age_range=value, on_date='CURRENT')

                queues['40'][value] = (
                    queues['38'][value] + queues['39'][value])

                queues['41'][value] = queue_index.get_count(
                    index_type=QueueIndex.CURRENT_DATE_DEFERRED,
                    age_range=value, on_date='CURRENT')

                queues['42'][value] = (
                    queues['18.1'][value] +
                    queues['7.1'][value])

                queues['17.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.SHORT_STAY, age_range=value,
                    on_date='CURRENT')

            for value in const.AGE_CATEGORIES_EIGHT.values():
                # Количество детей в группы оздоровительной и
                # компенсирующей направленности
                queues['13.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.CORRECT, age_range=value,
                    on_date='CURRENT')
                queues['14.1'][value] = queue_index.get_count(
                    index_type=QueueIndex.HEALTH, age_range=value,
                    on_date='CURRENT')

            for value in const.AGE_CATEGORIES_CUT.values():
                queues['17'][value] = queue_index.get_count(
                    index_type=QueueIndex.SHORT_STAY, age_range=value)

            # Количество детей в очереди имеющих право зачисления на основе
            # федеральных и региональных льгот.
            queues['11'][const.ALL] = queue_index.get_count(
                index_type=QueueIndex.PRIVILEGE,
                privilege_type=const.REGIONAL,
                on_date='CURRENT')
            queues['10'][const.ALL] = 0
            for value in const.AGE_CATEGORIES_FULL.values():
                queues['10'][const.ALL] += queues['9.1'][value]
            queues['10'][const.ALL] -= queues['11'][const.ALL]

            self._update_cache(queues, self.QUEUES)

            return queues

    def _collect_enrolled_data(self, dou, queues):

        enrolled = SortedDict([
            ('19', SortedDict()), ('19.1', SortedDict()),
            ('20.1', SortedDict()),
            ('20.2', SortedDict()), ('21', SortedDict()),
            ('22', SortedDict()), ('22.1', SortedDict()),
            ('23.1', SortedDict()),
            ('24.1', SortedDict()), ('25', SortedDict()),
            ('25.1', SortedDict()),
            ('26', SortedDict()), ('26.1', SortedDict()),
            ('27', SortedDict()), ('27.1', SortedDict()),
            ('28', SortedDict()), ('29', SortedDict()),
            ('30', SortedDict()), ('30.1', SortedDict())
        ])

        with session_scope() as session:
            index = EnrolledIndex(dou, session)
            application_index = ApplicationIndex(dou, session)

            for value in const.AGE_CATEGORIES_FULL.values():
                # Общая численность детей зачисленных в ДОО
                enrolled['19'][value] = index.get_count(age_range=value)
                enrolled['19.1'][value] = index.get_count(
                    age_range=value, on_date=datetime.date.today)

                enrolled['20.1'][value] = application_index.get_count(
                    index_type=ApplicationIndex.WANT_CHANGE_DOU,
                    age_range=value, on_date=datetime.date.today)
                enrolled['20.2'][value] = application_index.get_count(
                    index_type=ApplicationIndex.WANT_CHANGE_DOU_WITH_HN,
                    age_range=value, on_date=datetime.date.today)

                # Общая численность детей, зачисленных в группы для
                # детей с ограниченными возможностями здоровью
                enrolled['22'][value] = index.get_count(
                    index_type=EnrolledIndex.COUNT_DISABLED,
                    age_range=value)
                enrolled['22.1'][value] = index.get_count(
                    index_type=EnrolledIndex.COUNT_DISABLED,
                    age_range=value, on_date=datetime.date.today)

                # Общее количество детей, зачисленных в негос. ДОО
                enrolled['25'][value] = index.get_count(
                    age_range=value, dou_type=const.NOT_GOVERNMENT)
                enrolled['25.1'][value] = index.get_count(
                    age_range=value, dou_type=const.NOT_GOVERNMENT,
                    on_date=datetime.date.today)

                # Общее количество детей, зачисленных к ИП
                enrolled['26'][value] = index.get_count(
                    age_range=value, dou_type=const.IP)
                enrolled['26.1'][value] = index.get_count(
                    age_range=value, dou_type=const.IP,
                    on_date=datetime.date.today)

                # Группы кратковременного пребывания
                enrolled['27.1'][value] = index.get_count(
                    age_range=value,
                    index_type=EnrolledIndex.COUNT_SHORT_STAY,
                    on_date=datetime.date.today)
                enrolled['30'][value] = (
                    enrolled['19'][value] + queues['18'][value])
                enrolled['30.1'][value] = (
                    enrolled['19.1'][value] + queues['18.1'][value])

            for value in const.AGE_CATEGORIES_EIGHT.values():
                # Общее количество детей зачисленных в
                # оздоровительные и компенсирующие группы
                enrolled['23.1'][value] = index.get_count(
                    index_type=EnrolledIndex.COUNT_CORRECT,
                    age_range=value, on_date=datetime.date.today)
                enrolled['24.1'][value] = index.get_count(
                    index_type=EnrolledIndex.COUNT_HEALTH,
                    age_range=value, on_date=datetime.date.today)

            for value in const.AGE_CATEGORIES_CUT.values():
                enrolled['27'][value] = index.get_count(
                    age_range=value,
                    index_type=EnrolledIndex.COUNT_SHORT_STAY)
                enrolled['28'][value] = enrolled['29'][value] = 0
                for age in const.CUT_TO_FULL_CATEGORY_MAP[value]:
                    enrolled['28'][value] += enrolled['25'][age]
                    enrolled['29'][value] += enrolled['26'][age]

            # Общее количество детей зачисленных в семейные,
            # оздоровительные и компенсирующие группы
            enrolled['21'][const.ALL] = index.get_count(
                index_type=EnrolledIndex.COUNT_FAMILY,
                on_date=datetime.date.today)

        self._update_cache(enrolled, self.ENROLLED)

        return enrolled

    def _collect_capacities(self, dou):

        capacities = SortedDict([
            ('31.1', SortedDict()), ('32.1', SortedDict()),
            ('32.2', SortedDict()), ('32.3', SortedDict()),
            ('33.1', SortedDict()), ('34.1', SortedDict()),
            ('35.1', SortedDict())
        ])

        index = CapacitiesIndex(dou)

        for value in const.AGE_CATEGORIES_EIGHT.values():
            capacities['31.1'][value] = index.get_count(
                age_category=value)
            capacities['32.1'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.HEALTH_NEED)
            capacities['32.2'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.COMPENSATING)
            capacities['32.3'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.HEALTH)
            capacities['33.1'][value] = index.get_count(
                age_category=value,
                dou_type=const.NOT_GOVERNMENT)
            capacities['34.1'][value] = index.get_count(
                age_category=value,
                dou_type=const.IP)
            capacities['35.1'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.SHORT_STAY)

        self._update_cache(capacities, self.CAPACITIES)
        return capacities


class DocumentService(object):
    def __init__(
            self, mo, date, schedule_settings,
            unit_helper, report_form):

        self._mo = mo
        self._date = date.date()
        self._gisdo_settings = schedule_settings
        self._unit_helper = unit_helper
        self._report_form = report_form
        self._document = {}

    def _create_header(self):
        """
        Формирование данных для заголовка
        """

        self._document.update(
            report_date=self._date.isoformat(),
            send_time=datetime.datetime.now().isoformat(),
            system_name=self._gisdo_settings.system_name,
            version=settings.KINDER_VERSION,
            email=self._gisdo_settings.email,
            install_type=InstallationType.values.get(
                self._gisdo_settings.installation_type,
                InstallationType.values[InstallationType.REGIONAL]),
            vendor=u'ЗАО БАРС Груп')

    def _get_municipality_info(self):

        result = {
            'OKTMO': self._mo.octmo or ''
        }

        if self._mo.epgu_link:
            result.update({'epguLink': self._mo.epgu_link})

        if self._mo.rpgu_link:
            result.update({'rpguLink': self._mo.rpgu_link})

        return result

    @staticmethod
    def _get_unit_info(dou):

        return OrganizationInfo.get_info(dou)

    @staticmethod
    def _get_contact_info(dou):

        return ContactInfo.get_info(dou)

    @staticmethod
    def _get_groups(dou):

        return GroupsInfo.get_info(dou)

    @staticmethod
    def _get_additional_info(dou, model=UnitDocFiles):

        extensions = model.objects.filter(
            unit=dou).values_list('name', flat=True)

        if not extensions.exists():
            return [u'Нет']
        else:
            return extensions

    @staticmethod
    def _transform(counter):

        return CounterProxy.transform(counter)

    def _get_counters(self, dou):

        rows = ReportFormRow.objects.filter(
            unit=dou, report=self._report_form)

        counters = []
        for row in rows:
            counter = {}
            counter.update(json.loads(row.applications))
            counter.update(json.loads(row.queues))
            counter.update(json.loads(row.enrolled))
            counter.update(json.loads(row.capacities))
            counters.extend(DocumentService._transform(counter))

        return counters

    @staticmethod
    def _get_group_counters(dou):

        return Group.objects.filter(unit=dou, status=GroupStatusEnum.FACT).values('id')

    def _create_reports(self):

        reports = []

        # Признак наличия садов по которым можно
        # собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo()

        for dou in dou_list:
            report = {}
            report.update(self._get_municipality_info())
            report.update({
                'unit': DocumentService._get_unit_info(dou),
                'contact': DocumentService._get_contact_info(dou),
                'groups': DocumentService._get_groups(dou),
                'additionalEducations': DocumentService._get_additional_info(
                    dou),
                'features': DocumentService._get_additional_info(
                    dou, model=DopService),
                'counters': self._get_counters(dou),
                'group_counters': DocumentService._get_group_counters(dou)
            })

            reports.append(report)

        self._document.update({'reports': reports})

    def create(self):

        self._create_header()
        self._create_reports()

        return self._document


class CounterProxy(object):
    @staticmethod
    def transform(counter):

        result = []
        for index_id in counter:

            index = {
                'id': index_id,
                'indexes': []
            }

            for age_cat in counter[index_id]:
                index['indexes'].append({
                    'category': age_cat,
                    'count': counter[index_id][age_cat]
                })

            result.append(index)

        return result


class OVZInfoMixin(object):
    NOT_OVZ_GROUPS = u'Нет групп ОВЗ'
    DIFFERENT_OVZ_GROUPS = u'Комбинированные группы ОВЗ'
    OVZ_TYPES = {
        HNE.DEAFNESS: u'Нарушения слуха',
        HNE.HARDOFHEARTING: u'Нарушения слуха',
        HNE.SPEACH: u'Нарушения речи',
        HNE.PHONETICS: u'Нарушения речи',
        HNE.AMAUROSIS: u'Нарушения зрения',
        HNE.BLINDNESS: u'Нарушения зрения',
        HNE.BACKLIGHT: u'Задержка психического развития',
        HNE.BACKHARD: u'Задержка психического развития',
        HNE.BACK: u'Умственная отсталость',
        HNE.DISABLEMENT: u'Нарушения опорно-двигательного аппарата',
        HNE.INVALIDITY: u'Сложные дефекты',
        HNE.OTHER: u'Иные нарушения в развитии'
    }

    @classmethod
    def _get_group_ovz_type(cls, group):

        if group.health_need:
            ovz_type = cls.OVZ_TYPES.get(
                group.health_need.code,
                cls.OVZ_TYPES[HNE.OTHER])
        else:
            ovz_type = cls.OVZ_TYPES.get(HNE.OTHER)

        return ovz_type

    @classmethod
    def _get_dou_ovz_type(cls, dou):

        dou_groups = Group.objects.filter(
            unit=dou,
            status=GroupStatusEnum.FACT).select_related('health_need')

        # Если ни у одной группы нет потребности
        if dou_groups.filter(health_need__isnull=True).exists():
            ovz_type = cls.NOT_OVZ_GROUPS
        else:
            # Коды потребностей групп
            hn_codes = []
            for group in dou_groups:
                if group.health_need:
                    code = group.health_need.code
                else:
                    code = None
                hn_codes.append(code)
            # Преобразуем в set. Повторяющиеся элементы
            # исчезнут
            hn_codes = set(hn_codes)
            # Имеются различные типы потребностей
            if len(hn_codes) > 1:
                ovz_type = cls.DIFFERENT_OVZ_GROUPS
            elif len(hn_codes) == 0:
                ovz_type = cls.NOT_OVZ_GROUPS
            else:
                # Только один элемент
                code = hn_codes.pop()
                ovz_type = cls.OVZ_TYPES.get(
                    code, cls.OVZ_TYPES[HNE.OTHER])

        return ovz_type


class OrganizationInfo(OVZInfoMixin):
    ORGANIZATION_TYPES = {
        DouType.GOVERNMENT: u'Государственная',
        DouType.MUNICIPAL: u'Муниципальная',
        DouType.NOT_GOVERNMENT_WITH_LICENSE: u'Негосударственная',
        DouType.NOT_GOVERNMENT_WITHOUT_LICENSE: (
            u'Негосударственная без лицензии на осуществление '
            u'образовательной деятельности'),
        DouType.IP_WITH_LICENSE: u'Индивидуальный предприниматель (ИП)',
        DouType.IP_WITHOUT_LICENSE: (
            u'Индивидуальный предприниматель (ИП) без лицензии '
            u'на осуществление образовательной деятельности'),
        DouType.DEPARTMENTAL: u'Ведомственная',
        DouType.OTHER: u'Государственная'
    }

    BUILD_IN_PROGRESS = 1
    CLOSED = 2
    IN_CAPITAL_REPAIR = 3
    WORK = 4

    STATUSES = {
        BUILD_IN_PROGRESS: u'Ещё не открыто',
        CLOSED: u'Деятельность приостановлена',
        IN_CAPITAL_REPAIR: u'Капитальный ремонт',
        WORK: u'Функционирует'
    }

    @classmethod
    def _get_dou_type(cls, dou):

        dou_type = dou.dou_type

        if not dou_type:
            code = DouType.GOVERNMENT
        else:
            code = dou_type.code

        return cls.ORGANIZATION_TYPES[code]

    @classmethod
    def _get_dou_status(cls, dou):

        status = cls.STATUSES[cls.WORK]

        if dou.closed:
            status = cls.STATUSES[cls.CLOSED]
        elif dou.build_in_progress:
            status = cls.STATUSES[cls.BUILD_IN_PROGRESS]
        elif dou.in_capital_repair:
            status = cls.STATUSES[cls.IN_CAPITAL_REPAIR]

        return status

    @classmethod
    def get_info(cls, dou):

        result = {
            'code': dou.gisdo.doo_identity,
            'name': dou.name,
            'type': cls._get_dou_type(dou),
            'status': cls._get_dou_status(dou),
            'ovzType': cls._get_dou_ovz_type(dou),
            'director': dou.boss_fio,
        }

        if dou.boss_fio:
            result.update({'boss_fio': dou.boss_fio})

        return result


class ContactInfo(object):
    @classmethod
    def get_info(cls, dou):

        result = {}

        if dou.address_house_guid:
            result.update({'fiasHouseGuid': dou.address_house_guid})
        if dou.address_full:
            result.update({'plainAddress': dou.address_full})
        if dou.site:
            result.update({'website': dou.site})
        if dou.email:
            result.update({'email': dou.email})
        if dou.telephone:
            result.update({'phone': dou.telephone})

        return result


class GroupsInfo(OVZInfoMixin):
    @classmethod
    def _get_capacity(cls, group):

        return Group.objects.extra(select={
            'count_norm': """
                SELECT "{t1}"."count" FROM "{t1}"
                    WHERE ("{t1}"."age_cat_id" = {t2} and
                        "{t1}"."cathegory_id" = {t3} and
                        (CASE WHEN "{t1}"."health_need_id" IS NOT NULL
                        THEN "{t1}"."health_need_id"={t4}
                        ELSE 1=1 END)) limit 1
            """.format(
                t1=GroupCountNorm._meta.db_table,
                t2=group.age_cat_id or 'null',
                t3=group.cathegory_id or 'null',
                t4=group.health_need_id or 'null'),
        })[0].count_norm or 0

    @classmethod
    def _get_enrolled(cls, group):

        return Pupil.objects.filter(
            grup=group, temporary=False,
            temporary_deduct=False).count()

    @classmethod
    def get_info(cls, dou):

        dou_groups = Group.objects.filter(
            unit=dou, status=GroupStatusEnum.FACT)

        result = []
        for group in dou_groups:

            age_sub_category = group.sub_age_cat
            if age_sub_category:
                code = age_sub_category.code
                age_from, age_to = code.split('-')
            else:
                age_from, age_to = '1', '7'

            group_dict = {
                'id': group.id,
                'ageFrom': age_from,
                'ageTo': age_to,
                'ovzType': cls._get_group_ovz_type(group),
                'capacity': cls._get_capacity(group),
                'enrolled': cls._get_enrolled(group)
            }

            if group.name:
                group_dict.update(
                    {'name': group.name}
                )

            result.append(group_dict)

        return result


class TemplateLoader(object):
    TEMPLATE = 'ws/gisdo_request_template.xml'

    def __init__(self, schedule_settings):

        self._settings = schedule_settings

    def load(self, document):
        xml_string = loader.render_to_string(
            self.TEMPLATE, document).encode('utf-8')
        if self._settings.zip_xml:
            xml_string = zlib.compress(xml_string)[2:-4]

        encoded_xml = base64.b64encode(xml_string)

        return encoded_xml


class BroadCaster(object):
    def __init__(self, schedule_settings):

        self._settings = schedule_settings
        self._client = self._create_client()

    def _create_client(self):

        try:
            header = Element('ns1:Authentication')
            header.append(Element('ns1:Login').setText(
                self._settings.push_login))
            header.append(Element('ns1:Password').setText(
                self._settings.push_password))
            client = Client(
                url=gs_settings.GISDO_WS_WSDL,
                timeout=gs_settings.GISDO_TIMEOUT)
            client.set_options(soapheaders=(header,))
        except Exception:
            raise SOAPClientCreateException()
        return client

    def _get_send_method(self):

        if self._settings.zip_xml:
            if self._settings.async_send:
                method = 'PushZipDataAsync'
            else:
                method = 'PushZipData'
        else:
            method = 'PushData'

        return method

    def _get_number_of_attempts(self):

        return int(self._settings.resend_count)

    def send(self, content):

        method = self._get_send_method()

        attempts = 1
        error_msg = None
        while attempts < self._get_number_of_attempts():
            try:
                result = getattr(self._client.service, method)(
                    data=content, schema=u'2.0')
            except (socket.timeout, suds.WebFault) as error:
                error_msg = str(error).encode('utf-8')
                attempts += 1
            except Exception, error:
                error_msg = str(error).encode('utf-8')
                break
            else:
                if not result['Result']:
                    error_msg = result['Message']
                    break
                else:
                    break

        if error_msg:
            Logger.add_record(error_msg, level=Logger.ERROR)
            raise SendReportException()
